import itertools
import os
import sys
import traceback

import click
from termcolor import colored

from . import core, jncapi

DEBUG = False


def login_option(f):
    return click.option(
        "-l",
        "--email",
        required=True,
        envvar="JNCEP_EMAIL",
        help="Login email for J-Novel Club account",
    )(f)


def password_option(f):
    return click.option(
        "-p",
        "--password",
        required=True,
        envvar="JNCEP_PASSWORD",
        help="Login password for J-Novel Club account",
    )(f)


def output_option(f):
    return click.option(
        "-o",
        "--output",
        "output_dirpath",
        type=click.Path(exists=True, resolve_path=True, file_okay=False, writable=True),
        default=os.getcwd(),
        help="Existing folder to write the output [default: The current directory]",
    )(f)


def byvolume_option(f):
    return click.option(
        "-v",
        "--byvolume",
        "is_by_volume",
        is_flag=True,
        help=(
            "Flag to indicate that the parts of different volumes shoud be output in "
            "separate EPUBs"
        ),
    )(f)


def images_option(f):
    return click.option(
        "-i",
        "--images",
        "is_extract_images",
        is_flag=True,
        help=(
            "Flag to indicate that the images of the novel should be extracted into "
            "the output folder"
        ),
    )(f)


@click.group()
def cli():
    pass


@cli.command(name="epub", help="Generate EPUB files for J-Novel Club pre-pub novels")
@click.argument("jnc_url", metavar="JNOVEL_CLUB_URL", required=True)
@login_option
@password_option
@output_option
@click.option(
    "-s",
    "--parts",
    "part_specs",
    help=(
        "Specification of a range of parts to download in the form of "
        "<vol>[.part]:<vol>[.part] [default: All the content linked by "
        "the JNOVEL_CLUB_URL argument, either a single part, a whole volume "
        "or the whole series]"
    ),
)
@click.option(
    "-a",
    "--absolute",
    "is_absolute",
    is_flag=True,
    help=(
        "Flag to indicate that the --parts option specifies part numbers "
        "globally, instead of relative to a volume i.e. <part>:<part>"
    ),
)
@byvolume_option
@images_option
def generate_epub(
    jnc_url,
    email,
    password,
    part_specs,
    is_absolute,
    output_dirpath,
    is_by_volume,
    is_extract_images,
):
    slug = jncapi.slug_from_url(jnc_url)

    print(f"Login with email '{email}'...")
    token = jncapi.login(email, password)

    print(f"Fetching metadata for '{slug[0]}'...")
    metadata = jncapi.fetch_metadata(token, slug)

    novel = core.analyze_novel_metadata(slug[1], metadata)
    if part_specs:
        print(
            f"Using part specification '{part_specs}' "
            f"(absolute={_to_yn(is_absolute)})..."
        )
        parts_to_download = core.analyze_part_specs(novel, part_specs, is_absolute)
    else:
        parts_to_download = core.analyze_requested(novel)

    _create_epub_with_parts(
        token, novel, parts_to_download, is_by_volume, output_dirpath, is_extract_images
    )


@cli.command(name="track", help="Track updates to a series")
@click.argument("jnc_url", metavar="(JNOVEL_CLUB_URL?)", required=False)
@login_option
@password_option
@click.option(
    "--rm",
    "is_rm",
    is_flag=True,
    help="Flag to indicate that the series identified by the JNOVEL_CLUB_URL argument "
    "should be untracked",
)
def track_series(jnc_url, email, password, is_rm):
    if is_rm and not jnc_url:
        raise ValueError("A JNOVEL_CLUB_URL must be passed")

    if not jnc_url:
        # list
        tracked_series = core.read_tracked_series()
        if len(tracked_series) > 0:
            print(f"{len(tracked_series)} series are tracked:")
            for ser_url, ser_details in tracked_series.items():
                if isinstance(ser_details, dict):
                    print(f"'{ser_details.name}' ({ser_url}): {ser_details.part}")
                else:
                    # keep compat with old version for now
                    print(f"'{ser_url}': {ser_details}")
        else:
            print(f"No series is tracked.")

        return

    slug = jncapi.slug_from_url(jnc_url)

    print(f"Login with email '{email}'...")
    token = jncapi.login(email, password)

    print(f"Fetching metadata for '{slug[0]}'...")
    metadata = jncapi.fetch_metadata(token, slug)

    novel = core.analyze_novel_metadata(slug[1], metadata)
    # standardize on the series slug for the config (even though URLs
    # for volumes or parts are accepted)
    series_slug = novel.raw_serie.titleslug
    series_url = jncapi.url_from_slug(series_slug)

    tracked_series = core.read_tracked_series()

    if not is_rm:
        # keep compatibility for now
        if series_slug in tracked_series or series_url in tracked_series:
            print(
                colored(
                    f"The series '{novel.raw_serie.title}' is already tracked!",
                    "yellow",
                )
            )
            return

        # record current last part + name
        pn = core.to_relative_part_string(novel, novel.parts[-1])
        tracked_series[series_url] = {"part": pn, "name": novel.raw_serie.title}
        core.write_tracked_series(tracked_series)

        relative_part = core.to_relative_part_string(novel, novel.parts[-1])
        print(
            colored(
                f"The series '{novel.raw_serie.title}' is now tracked, starting after "
                f"part {relative_part}",
                "green",
            )
        )
    else:
        if series_slug not in tracked_series and series_url not in tracked_series:
            print(
                colored(
                    f"The series '{novel.raw_serie.title}' is not tracked!", "yellow"
                )
            )
            return

        # try both old and new form
        try:
            del tracked_series[series_slug]
        except Exception:
            del tracked_series[series_url]

        core.write_tracked_series(tracked_series)

        print(
            colored(
                f"The series '{novel.raw_serie.title}' is no longer tracked", "green"
            )
        )


@cli.command(
    name="update",
    help="Generate EPUB files for new parts of all tracked series (or specific "
    "series if a URL argument is passed)",
)
@click.argument("jnc_url", metavar="(JNOVEL_CLUB_URL?)", required=False)
@login_option
@password_option
@output_option
@byvolume_option
@images_option
def update_tracked(  # noqa: C901
    jnc_url, email, password, output_dirpath, is_by_volume, is_extract_images,
):

    tracked_series = core.read_tracked_series()
    if len(tracked_series) == 0:
        print(
            colored(
                "There are no tracked series! Use the 'jncep track' command first.",
                "yellow",
            )
        )
        return

    print(f"Login with email '{email}'...")
    token = jncapi.login(email, password)

    updated_series = []
    if jnc_url:
        slug = jncapi.slug_from_url(jnc_url)

        print(f"Fetching metadata for '{slug[0]}'...")
        metadata = jncapi.fetch_metadata(token, slug)

        novel = core.analyze_novel_metadata(slug[1], metadata)
        series_slug = novel.raw_serie.titleslug
        series_url = jncapi.url_from_slug(series_slug)

        if series_slug not in tracked_series and series_url not in tracked_series:
            print(
                colored(
                    f"The series '{novel.raw_serie.title}' is not tracked! "
                    f"Use the 'jncep track' command first.",
                    "yellow",
                )
            )
            return

        series_details = tracked_series[series_slug]
        # keep compatibility for now
        if isinstance(series_details, dict):
            last_pn = core.to_part(novel, series_details.part).raw_part.partNumber
        else:
            last_pn = series_details

        is_updated = _create_updated_epub(
            token, novel, last_pn, is_by_volume, output_dirpath, is_extract_images,
        )

        if is_updated:
            print(
                colored(
                    f"The series '{novel.raw_serie.title}' has been updated!", "green"
                )
            )
            updated_series.append(novel)
    else:
        # keep compatibility
        for series_slug_or_url, series_details in tracked_series.items():
            # see track command: always record the Novel slug
            try:
                slug = jncapi.slug_from_url(series_slug_or_url)
            except Exception:
                # not a URL. It is probably a slug from an older version
                slug = (series_slug_or_url, "NOVEL")
            print(f"Fetching metadata for '{slug[0]}'...")
            metadata = jncapi.fetch_metadata(token, slug)
            novel = core.analyze_novel_metadata(slug[1], metadata)

            # keep compatibility for now
            if isinstance(series_details, dict):
                last_pn = core.to_part(novel, series_details.part).raw_part.partNumber
            else:
                last_pn = series_details

            is_updated = _create_updated_epub(
                token, novel, last_pn, is_by_volume, output_dirpath, is_extract_images,
            )
            if is_updated:
                print(
                    colored(
                        f"The series '{novel.raw_serie.title}' has been updated!",
                        f"green",
                    )
                )
                updated_series.append(novel)

    if len(updated_series) > 0:
        # update tracking config JSON => to last part in series
        # TODO do that in the loop instead of the end ?
        for novel in updated_series:
            pn = core.to_relative_part_string(novel, novel.parts[-1])
            # write part + name in case old version with just the part number
            tracked_series[jncapi.url_from_slug(novel.raw_serie.titleslug)] = {
                "part": pn,
                "name": novel.raw_serie.title,
            }
            # wipeout old format if exists
            # TODO do that for all ?
            try:
                del tracked_series[novel.raw_serie.titleslug]
            except Exception:
                pass
        core.write_tracked_series(tracked_series)

        print(colored(f"{len(updated_series)} series sucessfully updated!", "green"))
    else:
        print(colored(f"All series are already up to date!", "green"))


def _to_yn(b):
    return "yes" if b else "no"


def _create_updated_epub(
    token, novel, last_pn, is_by_volume, output_dirpath, is_extract_images,
):
    if not novel.parts[-1].raw_part.partNumber > last_pn:
        # no new part
        print(
            colored(
                f"The series '{novel.raw_serie.title}' has not been updated!", "yellow",
            )
        )
        return False

    # create string part specs based on the next abs part number
    part_specs = f"{last_pn + 1}:"
    is_absolute = True
    parts_to_download = core.analyze_part_specs(novel, part_specs, is_absolute)

    _create_epub_with_parts(
        token,
        novel,
        parts_to_download,
        is_by_volume,
        output_dirpath,
        is_extract_images,
    )

    return True


def _create_epub_with_parts(
    token, novel, parts_to_download, is_by_volume, output_dirpath, is_extract_images
):
    # preview => parts 1 of each volume, always available
    # not expired => prepub
    available_parts_to_download = list(
        filter(
            lambda p: p.raw_part.preview or not p.raw_part.expired, parts_to_download
        )
    )

    if len(available_parts_to_download) == 0:
        raise NoRequestedPartAvailableError(
            "None of the requested parts are available for reading"
        )

    if len(available_parts_to_download) != len(parts_to_download):
        print(
            colored(
                "Some of the requested parts are not available for reading !", "yellow"
            )
        )

    if is_by_volume:
        for _, g in itertools.groupby(
            available_parts_to_download, lambda p: p.volume.volume_id
        ):
            parts = list(g)
            core.create_epub(token, novel, parts, output_dirpath, is_extract_images)
    else:
        core.create_epub(
            token, novel, available_parts_to_download, output_dirpath, is_extract_images
        )


class NoRequestedPartAvailableError(Exception):
    def __init__(self, msg):
        super().__init__(self, msg)


def main():
    try:
        cli()
    except Exception as ex:
        print(colored("*** An unrecoverable error occured ***", "red"))
        print(colored(str(ex), "red"))
        if DEBUG:
            traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    main()



from html.parser import HTMLParser
import json
import os
from pathlib import Path
import re
import time

from addict import Dict as Addict
from atomicwrites import atomic_write
import attr
from ebooklib import epub
from termcolor import colored

from . import jncapi

RANGE_SEP = ":"

CONFIG_DIRPATH = Path.home() / ".jncep"


@attr.s
class Novel:
    raw_serie = attr.ib()
    raw_metadata = attr.ib()
    requested_type = attr.ib()
    volumes = attr.ib(default=None)
    parts = attr.ib(default=None)


@attr.s
class Volume:
    raw_volume = attr.ib()
    volume_id = attr.ib()
    parts = attr.ib(factory=list)


@attr.s
class Part:
    raw_part = attr.ib()
    num_in_volume = attr.ib()
    volume = attr.ib()
    content = attr.ib(default=None)


def to_relative_part_string(novel, part):
    volume_number = part.volume.raw_volume.volumeNumber
    part_number = part.num_in_volume
    return f"{volume_number}.{part_number}"


def to_part(novel, relpart_str):
    # there will be an error if the relpart does not not existe
    parts = _analyze_volume_part_specs(novel, relpart_str)
    return parts[0]


def create_epub(token, novel, parts, output_dirpath, is_extract_images):
    contents, downloaded_img_urls = get_book_content_and_images(token, novel, parts)
    identifier, title, author, cover_url, toc = get_book_details(novel, parts)

    if cover_url in downloaded_img_urls:
        # no need to redownload
        # tuple : index 0 => bytes content
        # TODO do not add same file (same content, different name) in EPUB
        cover_bytes = downloaded_img_urls[cover_url][0]
    else:
        print("Fetching cover image...")
        cover_bytes = jncapi.fetch_image_from_cdn(cover_url)

    output_filename = _to_safe_filename(title) + ".epub"
    output_filepath = os.path.join(output_dirpath, output_filename)

    if is_extract_images:
        print("Extracting images...")
        # TODO better name than cloudfront URL
        for img_bytes, img_filename in downloaded_img_urls.values():
            img_filepath = os.path.join(output_dirpath, img_filename)
            with open(img_filepath, "wb") as img_f:
                img_f.write(img_bytes)

    create_epub_file(
        output_filepath,
        identifier,
        title,
        author,
        cover_bytes,
        parts,
        toc,
        contents,
        downloaded_img_urls,
    )
    print(colored(f"Success! EPUB generated in '{output_filepath}'!", "green"))


def get_book_content_and_images(token, novel, parts_to_download):
    downloaded_img_urls = {}
    contents = []
    for part in parts_to_download:
        print(f"Fetching part '{part.raw_part.title}'...")
        content = jncapi.fetch_content(token, part.raw_part.id)

        img_urls = _img_urls(content)
        if len(img_urls) > 0:
            print("Fetching images found in part content...")
            for i, img_url in enumerate(img_urls):
                print(f"Image {i + 1}...")
                # TODO catch, log  and ignore if error ?
                img_bytes = jncapi.fetch_image_from_cdn(img_url)
                # the filename relative to the epub content root
                # file will be added to the Epub archive
                # safe_filename on the base name (without the extension)
                new_local_filename = _to_safe_filename(img_url[:-4]) + img_url[-4:]
                downloaded_img_urls[img_url] = (img_bytes, new_local_filename)
                content = content.replace(img_url, new_local_filename)

        contents.append(content)

    return contents, downloaded_img_urls


def get_book_details(novel, parts_to_download):
    # shouldn't change between parts
    author = novel.raw_metadata.author
    if len(parts_to_download) == 1:
        # single part
        part = parts_to_download[0]
        identifier_base = part.raw_part.titleslug
        title = part.raw_part.title
        # use the first part of the volume: it contains a link to
        # a good resolution image
        cover_url = _cover_url(part.volume.parts[0].raw_part)
        # single part => single volume: part numbers relative to
        # that volume
        # TODO no TOC for single part ?
        toc = [f"Part {part.num_in_volume}"]
    else:
        volume_index = set()
        volumes = []
        for part in parts_to_download:
            if part.volume.volume_id in volume_index:
                continue
            volume_index.add(part.volume.volume_id)
            volumes.append(part.volume)

        if len(volumes) > 1:
            volume_nums = [str(volume.raw_volume.volumeNumber) for volume in volumes]
            volume_nums = ", ".join(volume_nums[:-1]) + " & " + volume_nums[-1]
            title_base = f"{novel.raw_serie.title}: Volumes {volume_nums}"

            part1 = to_relative_part_string(novel, parts_to_download[0])
            part2 = to_relative_part_string(novel, parts_to_download[-1])
            part_nums = f"Parts {part1} to {part2}"

            # TODO simplify instead ?
            toc = [part.raw_part.title for part in parts_to_download]

            # use the first part of the first volume in the requested content:
            # first part of a volume has cvr_860.jpg which is bigger than the cover_400
            # in volume or novel
            cover_url = _cover_url(volumes[0].parts[0].raw_part)
            title = f"{title_base} [{part_nums}]"
        else:
            volume = volumes[0]
            title_base = volume.raw_volume.title
            # same : use first part in volume which has cvr_860
            cover_url = _cover_url(volume.parts[0].raw_part)
            # relative to volume
            toc = [f"Part {part.num_in_volume}" for part in parts_to_download]

            # TODO totalPartNumber comes from the API and set only for some unfinished
            # volumes; If not set => maybe volume has all its parts ? if totalNumber
            # is there, not complete for sure but some unfinished volumes do not have it
            # either way
            # the volumes before the last are complete for sure
            if volume is not volumes[-1]:
                title = f"{title_base} [Complete]"
            else:
                title = (
                    f"{title_base} [Parts {parts_to_download[0].num_in_volume} to "
                    f"{parts_to_download[-1].num_in_volume}]"
                )

        identifier_base = novel.raw_serie.titleslug

    identifier = identifier_base + str(int(time.time()))

    return identifier, title, author, cover_url, toc


def create_epub_file(
    output_filepath,
    identifier,
    title,
    author,
    cover_bytes,
    parts,
    toc,
    contents,
    img_urls,
):
    lang = "en"
    book = epub.EpubBook()
    book.set_identifier(identifier)
    book.set_title(title)
    book.set_language(lang)
    book.add_author(author)

    book.set_cover("cover.jpg", cover_bytes, False)

    style = """body {color: black;}
h1 {page-break-before:always;}
img {width: 100%; page-break-after:always;page-break-before:always;}
.centerp {text-align: center;}
.noindent {text-indent: 0em;}"""
    css = epub.EpubItem(
        uid="style", file_name="book.css", media_type="text/css", content=style
    )
    book.add_item(css)

    cover_page = epub.EpubHtml(title="Cover", file_name="cover.xhtml", lang=lang)
    cover_page.content = '<img src="cover.jpg" alt="cover" />'
    cover_page.add_item(css)
    book.add_item(cover_page)

    for img_url, img_content in img_urls.items():
        img_bytes, local_filename = img_content
        img = epub.EpubImage()
        img.file_name = local_filename
        img.media_type = "image/jpeg"
        img.content = img_bytes
        book.add_item(img)

    chapters = []
    for i, content in enumerate(contents):
        c = epub.EpubHtml(title=toc[i], file_name=f"chap_{i +1}.xhtml", lang=lang)
        c.content = content
        c.add_item(css)
        book.add_item(c)
        chapters.append(c)

    book.toc = chapters

    book.add_item(epub.EpubNcx())
    book.add_item(epub.EpubNav())

    book.spine = [cover_page, "nav", *chapters]

    epub.write_epub(output_filepath, book, {})


def _cover_url(raw_metadata):
    covers = list(
        filter(
            lambda a: "cvr" in a.filename or "cover" in a.filename,
            raw_metadata.attachments,
        )
    )
    cover = max(covers, key=lambda c: c.size)
    return f"{jncapi.IMG_BASE_URL}{cover.fullpath}"


class ImgUrlParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.img_urls = []

    def handle_starttag(self, tag, tag_attrs):
        if tag != "img":
            return

        for tag_attr in tag_attrs:
            if tag_attr[0] == "src":
                self.img_urls.append(tag_attr[1])
                break


def _img_urls(content):
    parser = ImgUrlParser()
    parser.feed(content)
    return parser.img_urls


def _to_safe_filename(name):
    s = re.sub("[^0-9a-zA-Z]+", "_", name).strip("_")
    return s


def analyze_novel_metadata(req_type, metadata):
    """ Assumes the JSON returned from the API has all the parts and
    that they are ordered """
    # TODO don't trust and reorder no matter what ?

    if req_type in ("PART", "VOLUME"):
        novel = Novel(metadata.serie, metadata, req_type)
    else:
        novel = Novel(metadata, metadata, "NOVEL")

    volume_index = {}
    volumes = []
    for raw_volume in novel.raw_serie.volumes:
        volume = Volume(raw_volume, raw_volume.id)
        volumes.append(volume)
        volume_index[volume.volume_id] = volume

    parts = []
    for raw_part in novel.raw_serie.parts:
        volume_id = raw_part.volumeId
        volume = volume_index[volume_id]
        part = Part(raw_part, len(volume.parts) + 1, volume)
        volume.parts.append(part)
        parts.append(part)

    novel.volumes = volumes
    novel.parts = parts

    return novel


def analyze_requested(novel):
    if novel.requested_type == "PART":
        ip = novel.raw_metadata.partNumber - 1
        return [novel.parts[ip]]

    if novel.requested_type == "VOLUME":
        iv = novel.raw_metadata.volumeNumber - 1
        return list(novel.volumes[iv].parts)

    # novel: all parts
    return list(novel.parts)


def analyze_part_specs(novel, part_specs, is_absolute):
    """ v(.p):v2(.p) or v(.p): or :v(.p) or v(.p) or : """

    part_specs = part_specs.strip()

    if part_specs == RANGE_SEP:
        return novel.parts

    if is_absolute:
        return _analyze_absolute_part_specs(novel, part_specs)

    return _analyze_volume_part_specs(novel, part_specs)


def _analyze_absolute_part_specs(novel, part_specs):  # noqa: C901
    parts = []
    sides = part_specs.split(RANGE_SEP)
    if len(sides) > 2:
        raise ValueError("Multiple ':' in part specs")

    reg = r"^\s*(\d+)\s*$"
    if len(sides) == 1:
        # not a range: single part
        m = re.match(reg, sides[0])
        if not m:
            raise ValueError("Specified part must be a number")
        fp = int(m.group(1))
        ifp = _validate_absolute_part_number(novel, fp)
        return [novel.parts[ifp]]

    # range
    m1 = re.match(reg, sides[0])
    m2 = re.match(reg, sides[1])
    if not m1 and not m2:
        msg = "Part specification must be <number>:<number> or <number>: or :<number>"
        raise ValueError(msg)

    if m1:
        fp = int(m1.group(1))
        ifp = _validate_absolute_part_number(novel, fp)

    if m2:
        lp = int(m2.group(1))
        ilp = _validate_absolute_part_number(novel, lp)

    if m1 and not m2:
        # to the end
        for ip in range(ifp, len(novel.parts)):
            parts.append(novel.parts[ip])
        return parts

    if m2 and not m1:
        # since the beginning
        # + 1 => include the second side of the range
        for ip in range(0, ilp + 1):
            parts.append(novel.parts[ip])
        return parts

    # both sides are present
    if ifp > ilp:
        msg = "Second side of the part range must be greater than first"
        raise ValueError(msg)

    # + 1 => include the second side of the range
    for ip in range(ifp, ilp + 1):
        parts.append(novel.parts[ip])
    return parts


def _validate_absolute_part_number(novel, p):
    if p == 0:
        raise ValueError("Specified part number must be at least 1")
    # part specs start at 1 => transform to Python index
    ip = p - 1
    if ip >= len(novel.parts):
        raise ValueError(
            "Specified part number must be less than the number of parts in novel"
        )
    return ip


def _analyze_volume_part_specs(novel, part_specs):  # noqa: C901
    parts = []
    sides = part_specs.split(RANGE_SEP)
    if len(sides) > 2:
        raise ValueError("Multiple ':' in part specs")

    reg = r"^\s*(\d+)(?:\.(\d+))?\s*$"
    if len(sides) == 1:
        # not a range: single part
        m = re.match(reg, sides[0])
        if not m:
            raise ValueError(
                "Specification must be a of the form 'vol[.part]' (part is optional)"
            )
        fv = int(m.group(1))
        if m.group(2):
            # only the part specified
            fp = int(m.group(2))
            iv, ip = _validate_volume_part_number(novel, fv, fp)
            return [novel.volumes[iv].parts[ip]]
        else:
            # full volume
            iv = _validate_volume_part_number(novel, fv)
            for part in novel.volumes[iv].parts:
                parts.append(part)
            return parts

    # range
    m1 = re.match(reg, sides[0])
    m2 = re.match(reg, sides[1])
    if not m1 and not m2:
        msg = (
            "Part specification must be vol[.part]:vol[.part] or vol[.part]: or "
            ":vol[.part]"
        )
        raise ValueError(msg)

    if m1:
        fv = int(m1.group(1))
        if m1.group(2):
            fp = int(m1.group(2))
            ifv, ifp = _validate_volume_part_number(novel, fv, fp)
        else:
            ifv = _validate_volume_part_number(novel, fv)
            # beginning of the volume
            ifp = 0
        ifp = _to_absolute_part_index(novel, ifv, ifp)

    if m2:
        lv = int(m2.group(1))
        if m2.group(2):
            lp = int(m2.group(2))
            ilv, ilp = _validate_volume_part_number(novel, lv, lp)
        else:
            ilv = _validate_volume_part_number(novel, lv)
            # end of the volume
            ilp = -1
        # this works too if ilp == -1
        ilp = _to_absolute_part_index(novel, ilv, ilp)

    # same as for absolute part spec

    if m1 and not m2:
        # to the end
        for ip in range(ifp, len(novel.parts)):
            parts.append(novel.parts[ip])
        return parts

    if m2 and not m1:
        # since the beginning
        # + 1 => include the second side of the range
        for ip in range(0, ilp + 1):
            parts.append(novel.parts[ip])
        return parts

    # both sides are present
    if ifp > ilp:
        msg = "Second side of the vol[.part] range must be greater than first"
        raise ValueError(msg)

    # + 1 => always include the second side of the range
    for ip in range(ifp, ilp + 1):
        parts.append(novel.parts[ip])
    return parts


def _validate_volume_part_number(novel, v, p=None):
    iv = v - 1

    if iv >= len(novel.volumes):
        raise ValueError(
            "Specified volume number must be less than the number of volumes in novel"
        )
    volume = novel.volumes[iv]

    if p is None:
        return iv

    ip = p - 1
    if ip >= len(volume.parts):
        raise ValueError(
            "Specified part number must be less than the number of parts in volume"
        )
    return iv, ip


def _to_absolute_part_index(novel, iv, ip):
    volume = novel.volumes[iv]
    return volume.parts[ip].raw_part.partNumber - 1


def read_tracked_series():
    try:
        with _tracked_series_filepath().open() as json_file:
            data = json.load(json_file)
            return Addict(data)
    except FileNotFoundError:
        # first run ?
        return Addict({})


def write_tracked_series(tracked):
    _ensure_config_dirpath_exists()
    with atomic_write(_tracked_series_filepath().resolve(), overwrite=True) as f:
        f.write(json.dumps(tracked, sort_keys=True, indent=2))


def _tracked_series_filepath():
    return CONFIG_DIRPATH / "tracked.json"


def _ensure_config_dirpath_exists():
    CONFIG_DIRPATH.mkdir(parents=False, exist_ok=True)
